<?php

namespace Services\Lightspeed\api;

use GuzzleHttp;
use League\OAuth2\Client\Provider\GenericProvider;
use Psr\Http\Message\RequestInterface;

/**
 * This class is used to help authenticate with lightspeed retail api via oauth 2
 * and to authenticate requests / generate valid access tokens with refresh tokens
 * when the current access token expires.
 *
 * Class LightspeedOauth
 * @package Services\Lightspeed\api
 */
class LightspeedOauth
{
    const OAUTH2_AUTH_URL = 'https://cloud.lightspeedapp.com/oauth/authorize.php';
    const OAUTH2_TOKEN_URL = 'https://cloud.lightspeedapp.com/oauth/access_token.php';
    const API_BASE_URI = 'https://api.merchantos.com/API/';

    private $httpClient;
    private $token;
    private $scopes;
    private $config;
    private $accessTokenRefreshedCallback;

    /**
     * LightspeedClient constructor.
     * @param array $config
     */
    public function __construct($config = [])
    {
        $this->config = array_merge([
            'client_id' => 'TYd7SHxxR5ioPnpx',
            'client_secret' => '8FjjyhKF3XrwPSMh5cbN',
            'redirect_uri' => 'https://localhost:8000/oauth',
            'url_authorize' => self::OAUTH2_AUTH_URL,
            'url_access_token' => self::OAUTH2_TOKEN_URL,
            'scopes' => []
        ], $config);
        $this->scopes = [];
    }

    /**
     * Gets the array value of the access token object
     * @return array
     * @throws \League\OAuth2\Client\Provider\Exception\IdentityProviderException
     */
    public function getAccessToken(): array
    {
        return $this->token;
    }

    /**
     * Returns the "bearer" header for requests and checks to make sure
     * the access token is valid. If it is not valid then it requests a new
     * access token.
     */
    public function getBearer() {
        $auth = $this->makeOauth2Provider();
        if (!array_key_exists('expires', $this->token))
            $this->token['expires'] = 0;
        if (!self::isValid($this->token['expires'])) {
            $newToken = $auth->getAccessToken('refresh_token', [
                'refresh_token' => $this->token['refresh_token']
            ]);
            $tokenData = $this->token;
            $tokenData['access_token'] = $newToken->getToken();
            $tokenData['expires'] = $newToken->getExpires();
            $this->token = $tokenData;
            if ($this->accessTokenRefreshedCallback)
                ($this->accessTokenRefreshedCallback)($this->token);
        }
        return 'Bearer ' . $this->token['access_token'];
    }

    /**
     * @return array
     */
    public function authenticate(): array
    {
        $auth = $this->makeOauth2Provider();
        $authUrl = $auth->getAuthorizationUrl([
            'scopeSeparator' => '+',
            'scope' => $this->config['scopes']
        ]);
        return [
            'state' => $auth->getState(),
            'url' => $authUrl
        ];
    }

    /**
     * Generates the authentication url that a user uses to authenticate the
     * application with lightspeed.
     * @return string
     */
    public function getAuthenticationUrl(): string
    {
        $auth = $this->makeOauth2Provider();
        $authUrl = $auth->getAuthorizationUrl([
            'scopeSeparator' => '+',
            'scope' => $this->config['scopes']
        ]);

        return $authUrl;
    }

    /**
     * Add scopes for authentication
     * @param $scope
     */
    public function addScope($scope)
    {
        if (\is_array($scope)) {
            foreach ($scope as $s)
                $this->addScope($s);
        } else {
            $this->scopes[] = $scope;
        }
    }


    /**
     * Creates an authenticated request, and checks to make sure
     * that the acess token is valid. It will request a new access token if it's not.
     *
     * @param $method the request type
     * @param $url the url to call
     * @param array $params the query parameters
     * @param array $body the request body
     * @param array $options guzzle options
     * @return RequestInterface
     * @throws \League\OAuth2\Client\Provider\Exception\IdentityProviderException
     */
    public function createAuthenticatedRequest($method, $url, $params = [], $body = [], $options = [])
    {
        $auth = $this->makeOauth2Provider();

        if (!self::isValid($this->token['expires'])) {
            $newToken = $auth->getAccessToken('refresh_token', [
                'refresh_token' => $this->token['refresh_token']
            ]);
            $tokenData = $this->token;
            $tokenData['access_token'] = $newToken->getToken();
            $tokenData['expires'] = $newToken->getExpires();
            $this->token = $tokenData;
        }
        $options = [
            'debug' => true
        ];
        if (!empty($params))
            $options['query'] = $params;
        if (!empty($body))
            $options['json'] = $body;

        return $auth->getAuthenticatedRequest($method, $url, $this->token->getToken, $options);
    }

    /**
     * Retrieves the access token when provided with a temporary token
     * @param $code the temporary token
     * @return array the retrieved access token
     * @throws \League\OAuth2\Client\Provider\Exception\IdentityProviderException
     */
    public function retrieveAccessTokenWithCode($code): array
    {
        $auth = $this->makeOauth2Provider();
        $accessToken = $auth->getAccessToken('authorization_code', [
            'code' => $code,
            'grant_type' => 'authorization_code'
        ]);
        $creds = $accessToken->jsonSerialize();
        $creds['created'] = time();
        $this->setAccessToken($creds);
        return $this->token;
    }

    /**
     * Retrieves a new access token with the saved refresh_token
     * used when an access token has expired.
     *
     * @param $refresh_token
     * @return array
     */
    public function retrieveAccessTokenWithRefreshToken($refresh_token): array
    {
        $respBody = $this->refreshAccessToken($refresh_token);
        $tokenData = json_decode($respBody, true);
        $tokenData["timestamp"] = time();
        $tokenData["refresh_token"] = $refresh_token; // response doesn't contain the refresh token!
        $tokenData['created'] = time();
        $this->setAccessToken($tokenData);
        return $tokenData;
    }

    /**
     * @param array $accessToken
     */
    public function setAccessToken($accessToken = []): void
    {
        $this->token = $accessToken;
    }

    /**
     * @return bool
     */
    public function isAccessTokenExpired(): bool
    {
        if (!$this->token)
            return true;
        $created = 0;
        if (isset($this->token['created']))
            $created = $this->token['created'];

        return ($created + ($this->token['expires_in'] - 30)) < time();
    }

    /**
     * @param \Closure $callback
     */
    public function setAccessTokenRefreshedCB(\Closure $callback) {
        $this->accessTokenRefreshedCallback = $callback;
    }

    /**
     * Builds a oauth2 provided to handle the authentication flow and generate
     * neccessary urls
     * @return GenericProvider
     */
    protected function makeOauth2Provider(): GenericProvider
    {
        return new GenericProvider([
            'clientId' => $this->config['client_id'],
            'clientSecret' => $this->config['client_secret'],
            'redirectUri' => $this->config['redirect_uri'],
            'urlAuthorize' => $this->config['url_authorize'],
            'urlAccessToken' => $this->config['url_access_token'],
            'urlResourceOwnerDetails' => '',
        ]);
    }

    /**
     * @param $refresh_token
     * @return string
     */
    private function refreshAccessToken($refresh_token)
    {
        $query = [
            'refresh_token' => $refresh_token,
            'client_id' => $this->config['client_id'],
            'client_secret' => $this->config['client_secret'],
            'grant_type' => 'refresh_token'
        ];
        $response = $this->httpClient->post($this->config['url_access_token'], [
            'form_params' => $query
        ]);
        return $response->getBody()->getContents();
    }


    /**
     * @param $url
     * @param $query
     * @return mixed|\Psr\Http\Message\ResponseInterface
     * @throws GuzzleHttp\Exception\GuzzleException
     */
    private function makeGetRequest($url, $query)
    {

        $response = $this->httpClient->request('POST', $url, [
            'form_params' => $query
        ]);

        return $response;
    }


    private function makePostRequest($uri, $body = null): GuzzleHttp\Psr7\Request
    {
        $headers = [];
        return new GuzzleHttp\Psr7\Request('POST', $uri, $headers, $body);
    }

    /**
     * @param $timestamp
     * @return bool
     */
    private static function isValid($timestamp)
    {
        // check if token needs to be refreshed
        if ($timestamp <= strtotime("now")) {
            return false;
        }
        return true;
    }

}
