<?php


namespace App\Database;


use App\Database\Util\MysqlConverter;
use App\Models\Contact;
use App\Models\ContactAddress;
use App\Models\ContactEmail;
use App\Models\ContactPhone;
use App\Models\Customer;
use App\Models\CustomFieldValues;
use App\Models\Employee;
use App\Models\EmployeeHours;
use App\Models\InventoryCount;
use App\Models\InventoryCountCalc;
use App\Models\InventoryCountItem;
use App\Models\InventoryCountReconcile;
use App\Models\Item;
use App\Models\ItemComponent;
use App\Models\ItemPrice;
use App\Models\ItemShop;
use App\Models\ItemVendorNum;
use App\Models\Note;
use App\Models\Order;
use App\Models\OrderLine;
use App\Models\Sale;
use App\Models\SaleLine;
use App\Models\SalePayment;
use App\Models\Serialized;
use App\Models\Shop;
use App\Models\Workorder;
use App\Models\WorkorderItem;
use App\Models\WorkorderLine;
use App\Models\WorkorderStatus;
use Carbon\Carbon;
use GuzzleHttp\Exception\ClientException;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Storage;
use Services\Lightspeed\api\CustomerApi;
use Services\Lightspeed\api\EmployeeApi;
use Services\Lightspeed\api\InventoryApi;
use Services\Lightspeed\api\LightspeedOauth;
use Services\Lightspeed\api\RestClient;
use Services\Lightspeed\api\SaleApi;
use Services\Zoho\ZohoApi;

class LightspeedDataImporter
{
    use LightspeedPropHelpers;

    private $provider;
    private $workorderApi;
    private $customerApi;
    private $employeeApi;
    private $saleApi;
    private $inventoryApi;
    private $lsClient;

    /**
     * SyncLightspeedData constructor.
     */
    public function __construct()
    {

        $this->provider = App::make(LightspeedOauth::class);
        $this->provider->setAccessToken(json_decode(Storage::get('access_token.json'), true));
        $this->provider->setAccessTokenRefreshedCB(function ($accessToken) {
            Storage::put('access_token.json', json_encode($accessToken));
        });
        $this->workorderApi = new \Services\Lightspeed\api\WorkorderApi($this->provider, Storage::get('accountid'));
        $this->customerApi = new CustomerApi($this->provider, Storage::get('accountid'));
        $this->employeeApi = new EmployeeApi($this->provider, Storage::get('accountid'));
        $this->saleApi = new SaleApi($this->provider, Storage::get('accountid'));
        $this->inventoryApi = new InventoryApi($this->provider, Storage::get('accountid'));
        $this->lsClient = new RestClient($this->provider, Storage::get('accountid'));
    }


    private function getModifiedCheckDate($minutes = 10)
    {
        $timeStamp = Carbon::now()->subMinutes($minutes);
        return $timeStamp->format(\DateTime::ISO8601);
    }

    private function getLastModParams($timeStamp)
    {
        return ['timeStamp' => '>=,' . $timeStamp];
    }

    public function importWorkorders($update = false, $minutes = 10, $params = [])
    {
        $relations =  array_merge(['load_relations' => '["CustomFieldValues","Customer","Employee","Serialized","WorkorderStatus","WorkorderLines","WorkorderItems","CustomFieldValues"]', 'archived' => 'true'], $params);
        $this->lsClient->walkAll('Workorder', function ($workorders) {
            foreach ($workorders as $workorder) {
                if (property_exists($workorder, 'timeDone')) {
                    error_log('timeDone');
                }
                try {
                    if (config('database.default') == 'mysql') {
                        MysqlConverter::checkAndConvert($workorder);
                    }
                    Workorder::attrUpdateOrCreate('workorderID', Workorder::getFillableFields(), $workorder);
                    if (property_exists($workorder, 'WorkorderItems')) {
                        $woItems = $this->getRelated($workorder->WorkorderItems->{'WorkorderItem'});
                        foreach ($woItems as $i) {
                            if (config('database.default') == 'mysql') {
                                MysqlConverter::checkAndConvert($i);
                            }
                            WorkorderItem::attrUpdateOrCreate('workorderItemID', WorkorderItem::getFillableFields(), (array)$i);
                        }
                    }
                    if (property_exists($workorder, 'WorkorderLines')) {
                        $woLines = $this->getRelated($workorder->WorkorderLines->{'WorkorderLine'});
                        foreach ($woLines as $line) {
                            if (config('database.default') == 'mysql') {
                                MysqlConverter::checkAndConvert($line);
                            }
                            WorkorderLine::attrUpdateOrCreate('workorderLineID', WorkorderLine::getFillableFields(), (array)$line);
                        }
                    }
                } catch (\Exception $e) {
                    Log::error("Well this is odd");
                    Log::error($e->getMessage());
                    Log::error($e->getTraceAsString());
                }
            }
        }, $update ? array_merge($this->getLastModParams($this->getModifiedCheckDate($minutes)), $relations) : $relations);
    }

    public function importWorkorderStatuses($update = false, $minutes = 10)
    {
        $status = $this->workorderApi->getWorkorderStatuses();
        foreach ($status->WorkorderStatus as $s) {
            WorkorderStatus::attrUpdateOrCreate("workorderStatusID", WorkorderStatus::getFillableFields(), (array)$s);
        }
    }

    public function importSales($update = false, $minutes = 10)
    {
        $params = $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : [];
        $this->lsClient->walkAll('Sale', function ($sales) {
            foreach ($sales as $sale) {
                if (config('database.default') == 'mysql') {
                    MysqlConverter::checkAndConvert($sale);
                }
                Sale::attrUpdateOrCreate('saleID', Sale::getFillableFields(), (array)$sale);
                if (property_exists($sale, 'SaleLines')) {
                    foreach ($this->getRelated($sale->SaleLines->SaleLine) as $sl) {
                        if (config('database.default') == 'mysql') {
                            MysqlConverter::checkAndConvert($sl);
                        }
                        SaleLine::attrUpdateOrCreate('saleLineID', SaleLine::getFillableFields(), (array)$sl);
                    }
                }

                if (property_exists($sale, 'SalePayments')) {
                    foreach ($this->getRelated($sale->SalePayments->SalePayment) as $sp) {
                        if (config('database.default') == 'mysql') {
                            MysqlConverter::checkAndConvert($sp);
                        }
                        SalePayment::attrUpdateOrCreate('salePaymentID', SalePayment::getFillableFields(), (array)$sp);
                    }
                }
            }
        }, $params);
    }

    public function importShops($update = false, $minutes = 10)
    {
        $params = $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : [];
        $this->lsClient->walkAll('Shop', function ($shops) {
            foreach ($shops as $shp) {
                Shop::attrUpdateOrCreate('shopID', Shop::getFillableFields(), (array)$shp);
            }
        }, $params);
    }

    public function importOrders($update = false, $minutes = 10)
    {
        $params = $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : [];
        $this->lsClient->walkAll('Order', function ($orders) {
            foreach ($orders as $order) {
                MysqlConverter::checkAndConvert($order);
                Order::attrUpdateOrCreate('orderID', Order::getFillableFields(), (array)$order);
                if (property_exists($order, 'OrderLines')) {
                    foreach ($this->getRelated($order->OrderLines->OrderLine) as $ol) {
                        MysqlConverter::checkAndConvert($ol);
                        OrderLine::attrUpdateOrCreate('orderLineID', OrderLine::getFillableFields(), (array)$ol);
                    }
                }
            }
        }, $params);
    }

    public function importItems($update = false, $minutes = 10)
    {
        $params = $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : [];
        $this->lsClient->walkAll('Item', function ($items) {
            foreach ($items as $i) {
                MysqlConverter::checkAndConvert($i);
                Item::attrUpdateOrCreate('itemID', Item::getFillableFields(), (array)$i);

                if (property_exists($i, 'ItemShops')) {
                    foreach ($this->getRelated($i->ItemShops->ItemShop) as $is) {
                        ItemShop::attrUpdateOrCreate('itemShopID', ItemShop::getFillableFields(), (array)$is);
                    }
                }
                if (property_exists($i, 'ItemComponents')) {
                    foreach ($this->getRelated($i->ItemComponents->ItemComponent) as $ic) {
                        ItemComponent::attrUpdateOrCreate('itemComponentID', ItemComponent::getFillableFields(), (array)$ic);
                    }
                }

                if (property_exists($i, 'ItemVendorNums')) {
                    foreach ($this->getRelated($i->ItemVendorNums->ItemVendorNum) as $ivn) {
                        ItemVendorNum::attrUpdateOrCreate('itemVendorNumID', ItemVendorNum::getFillableFields(), $ivn);
                    }
                }

                if (property_exists($i, 'Prices')) {
                    foreach ($this->getRelated($i->Prices->ItemPrice) as $ip) {
                        ItemPrice::attrUpdateOrCreate([['itemID' => $i->itemID], 'useType'], ItemPrice::getFillableFields(), $ip);
                    }
                }
            }
        }, $params);
    }

    public function importEmployees($update = false, $minutes = 10, $params = [])
    {
        $this->lsClient->walkAll('Employee', function ($employees) {
            foreach ($employees as $emp) {
                MysqlConverter::checkAndConvert($emp);
                Employee::attrUpdateOrCreate('employeeID', Employee::getFillableFields(), $emp);
            }
        }, array_merge($update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : [], $params));
    }

    public function importEmployeeHours($update = false, $field = 'checkOut', $minutes = 10)
    {
        $params = $update ? [$field => '>=,' . $this->getModifiedCheckDate($minutes)] : [];
        try {
            $this->lsClient->walkAll('EmployeeHours', function ($employeeHours) {
                foreach ($employeeHours as $employeeHour) {
                    EmployeeHours::attrUpdateOrCreate('employeeHoursID', EmployeeHours::getFillableFields(), $employeeHour);
                }
            }, $params);
        } catch (\Exception $e) {
            Log::error($e);
            if ($e instanceof ClientException) {
                Log::error($e->getResponse()->getBody());
            }
        }
    }

    public function importCustomers($update = false, $minutes = 10)
    {
        $params = $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : [];
        $this->lsClient->walkAll('Customer', function ($customers) {
            foreach ($customers as $cust) {
                MysqlConverter::checkAndConvert($cust, true);
                Customer::attrUpdateOrCreate('customerID', Customer::getFillableFields(), $cust);
                if (property_exists($cust, 'Note')) {
                    Note::attrUpsert(array_merge((array)$cust->Note, ['objectID' => $cust->customerID, 'objectType' => 'Customer']));
                }

                if (property_exists($cust, 'Contact')) {
                    Contact::attrUpsert(array_merge((array)$cust->Contact, ['customerID' => $cust->customerID]));
                    if (property_exists($cust->Contact, 'Addresses') && !empty($cust->Contact->Addresses->ContactAddress)) {
                        foreach ($this->getRelated($cust->Contact->Addresses->ContactAddress) as $ca)
                            ContactAddress::attrUpsert(array_merge((array)$ca, ['customerID' => $cust->customerID, 'contactID' => $cust->Contact->contactID]));
                    }
                    if (property_exists($cust->Contact, 'Phones') && !empty($cust->Contact->Phones->ContactPhone)) {
                        foreach ($this->getRelated($cust->Contact->Phones->ContactPhone) as $cp)
                            ContactPhone::attrUpsert(array_merge((array)$cp, ['customerID' => $cust->customerID, 'contactID' => $cust->Contact->contactID]));
                    }
                    if (property_exists($cust->Contact, 'Emails') && !empty($cust->Contact->Emails->ContactEmail)) {
                        foreach ($this->getRelated($cust->Contact->Emails->ContactEmail) as $ce)
                            ContactEmail::attrUpsert(array_merge((array)$ce, ['customerID' => $cust->customerID, 'contactID' => $cust->Contact->contactID]));
                    }
                }
            }
        }, $params);
    }

    public function importInventory($update = false, $minutes = 10)
    {
        $this->lsClient->walkAll('InventoryCount', function ($inventoryCounts) {
            foreach ($inventoryCounts as $inventoryCount) {
                InventoryCount::attrUpdateOrCreate('inventoryCountID', InventoryCount::getFillableFields(), $inventoryCount);
            }
        }, $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : []);

        $this->lsClient->walkAll('InventoryCountItem', function ($inventoryCountItems) {
            foreach ($inventoryCountItems as $inventoryCountItem) {
                InventoryCountItem::attrUpdateOrCreate('inventoryCountItemID', InventoryCountItem::getFillableFields(), $inventoryCountItem);
            }
        }, $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : []);

        $this->lsClient->walkAll('InventoryCountCalc', function ($inventoryCountItems) {
            foreach ($inventoryCountItems as $inventoryCountItem) {
                InventoryCountCalc::attrUpdateOrCreate('inventoryCountCalcID', InventoryCountCalc::getFillableFields(), $inventoryCountItem);
            }
        }, $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : []);

        $this->lsClient->walkAll('InventoryCountReconcile', function ($inventoryReconciles) {
            foreach ($inventoryReconciles as $icr) {
                InventoryCountReconcile::attrUpdateOrCreate('inventoryCountReconcileID', InventoryCountReconcile::getFillableFields(), $icr);
            }
        }, $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : []);
    }

    public function importSerialized($update, $minutes = 10) {
        $this->lsClient->walkAll('Serialized', function ($serialized) {
            foreach ($serialized as $s) {
                Serialized::attrUpdateOrCreate('serializedID', Serialized::getFillableFields(), $s);
            }
        }, $update ? $this->getLastModParams($this->getModifiedCheckDate($minutes)) : []);
    }
}
